#ifndef RROOTS_OO_1_H
#define RROOTS_OO_1_H

#include <boost/python.hpp>

#ifdef _WIN32

#define WINVER 0x0601
#define _WIN32_WINNT 0x0601

#ifndef M_PI
#define M_PI 3.14159265358979323846264338328
#endif

/*
 * Obviously, because the number written above is only a
 * 'double' Literal, the full series of 30 decimal places
 * will not get used. The equivalent precision is to
 * 14 decimal places.
 *
 */

#endif

// Edit here.

#include <cmath>
#include <complex>
#include <algorithm>

#define MIN_A0 0.01


//#ifndef max
//#define max(a, b)  ((a > b) ? a : b)
//#endif


using std::complex;
using namespace std::complex_literals;

typedef complex<long double> CC;

struct Polynomial {
    CC *coeffs;
    int deg;
    int offset;
    CC *roots;
    int solved;
};


class PolySolver
{

public:
    PolySolver(boost::python::list coeffs);
    ~PolySolver();

    int poly_solve(bool polish, bool reverse,
                    boost::python::list sol_list);

    template<class T>
    static T abs_d(T x);

    double at(double x);

private:
    void normalize(Polynomial *a);
    static CC apply_poly(Polynomial *a, CC param);
    int divide(Polynomial *a, CC root, CC *remaind);

    bool try_root(Polynomial *a, CC *root, bool polish,
                  long double epsilon = 1.0E-12);

    inline bool solve_lin(Polynomial *a, CC *root);

    bool search_root(Polynomial *a, CC *root,
                     long double epsilon = 1.0E-12);

    int create_deriv_list(Polynomial a, Polynomial **l);
    void destroy_deriv_list(Polynomial *l);

    bool climb_root(Polynomial *L, CC *root, int i, int mult,
                    long double epsilon = 1.0E-12);

    void sort_roots(Polynomial *solved_p);

    void reverse_factor(Polynomial *target_p, Polynomial *solved_p,
                        long double epsilon = 1.0E-11);


    Polynomial *PL;
    Polynomial A;

};


// Define absolute of real number.
// Should not be necessary.
template<class T>
T PolySolver::abs_d(T x)
{
    if (x < 0) return -x;
    return x;
}


#endif
